@@ -1,7 +1,6 @@
 ;;; flyspell.el --- on-the-fly spell checker
 
-;; Copyright (C) 1998, 2000, 2001, 2002, 2003, 2004,
-;;   2005, 2006, 2007, 2008 Free Software Foundation, Inc.
+;; Copyright (C) 1998, 2000-2011  Free Software Foundation, Inc.
 
 ;; Author: Manuel Serrano <Manuel.Serrano@sophia.inria.fr>
 ;; Maintainer: FSF
@@ -72,13 +71,23 @@ Detection of repeated words is not implemented in
   :type 'boolean)
 
 (defcustom flyspell-mark-duplications-exceptions
-  '(("francais" . ("nous" "vous")))
+  '((nil . ("that" "had")) ; Common defaults for English.
+    ("\\`francais" . ("nous" "vous")))
   "A list of exceptions for duplicated words.
-It should be a list of (LANGUAGE . EXCEPTION-LIST).  LANGUAGE is matched
-against the current dictionary and EXCEPTION-LIST is a list of strings.
-The duplicated word is downcased before it is compared with the exceptions."
+It should be a list of (LANGUAGE . EXCEPTION-LIST).
+
+LANGUAGE is nil, which means the exceptions apply regardless of
+the current dictionary, or a regular expression matching the
+dictionary name (`ispell-local-dictionary' or
+`ispell-dictionary') for which the exceptions should apply.
+
+EXCEPTION-LIST is a list of strings.  The checked word is
+downcased before comparing with these exceptions."
   :group 'flyspell
-  :type '(alist :key-type string :value-type (repeat string)))
+  :type '(alist :key-type (choice (const :tag "All dictionaries" nil)
+				  string)
+		:value-type (repeat string))
+  :version "24.1")
 
 (defcustom flyspell-sort-corrections nil
   "Non-nil means, sort the corrections alphabetically before popping them."
@@ -199,9 +208,9 @@ Ispell's ultimate default dictionary."
 
 (defcustom flyspell-check-tex-math-command nil
   "Non-nil means check even inside TeX math environment.
-TeX math environments are discovered by the TEXMATHP that implemented
-inside the texmathp.el Emacs package.  That package may be found at:
-http://strw.leidenuniv.nl/~dominik/Tools"
+TeX math environments are discovered by `texmathp', implemented
+inside AUCTeX package.  That package may be found at
+URL `http://www.gnu.org/software/auctex/'"
   :group 'flyspell
   :type 'boolean)
 
@@ -242,7 +251,7 @@ spawned for speed.
 Doubled words are not detected in a large region, because Ispell
 does not check for them.
 
-If `flyspell-large-region' is nil, all regions are treated as small."
+If this variable is nil, all regions are treated as small."
   :group 'flyspell
   :version "21.1"
   :type '(choice number (const :tag "All small" nil)))
@@ -262,10 +271,17 @@ If `flyspell-large-region' is nil, all regions are treated as small."
   :group 'flyspell
   :type '(choice string (const nil)))
 
+(defvar flyspell-mode-map)
+
 (defcustom flyspell-use-meta-tab t
   "Non-nil means that flyspell uses M-TAB to correct word."
   :group 'flyspell
-  :type 'boolean)
+  :type 'boolean
+  :initialize 'custom-initialize-default
+  :set (lambda (sym val)
+	 (define-key flyspell-mode-map "\M-\t"
+	   (if (set sym val)
+	       'flyspell-auto-correct-word))))
 
 (defcustom flyspell-auto-correct-binding
   [(control ?\;)]
@@ -280,7 +296,7 @@ If `flyspell-large-region' is nil, all regions are treated as small."
 ;;*    using flyspell with mail-mode add the following expression       */
 ;;*    in your .emacs file:                                             */
 ;;*       (add-hook 'mail-mode                                          */
-;;*    	     '(lambda () (setq flyspell-generic-check-word-predicate    */
+;;*    	     (lambda () (setq flyspell-generic-check-word-predicate    */
 ;;*    			       'mail-mode-flyspell-verify)))            */
 ;;*---------------------------------------------------------------------*/
 (defvar flyspell-generic-check-word-predicate nil
@@ -340,40 +356,29 @@ property of the major mode name.")
    (not (save-excursion
 	  (re-search-backward "^[ \t]*%%%[ \t]+Local" nil t)))
    (not (save-excursion
-	  (let ((this (point-marker))
-		(e (progn (end-of-line) (point-marker))))
+	  (let ((this (point)))
 	    (beginning-of-line)
-	    (if (re-search-forward "\\\\\\(cite\\|label\\|ref\\){[^}]*}" e t)
-		(and (>= this (match-beginning 0))
-		     (<= this (match-end 0)) )))))))
+	    (and (re-search-forward "\\\\\\(cite\\|label\\|ref\\){[^}]*}"
+				    (line-end-position) t)
+		 (>= this (match-beginning 0))
+		 (<= this (match-end 0))))))))
 
 ;;*--- sgml mode -------------------------------------------------------*/
 (put 'sgml-mode 'flyspell-mode-predicate 'sgml-mode-flyspell-verify)
 (put 'html-mode 'flyspell-mode-predicate 'sgml-mode-flyspell-verify)
 (put 'nxml-mode 'flyspell-mode-predicate 'sgml-mode-flyspell-verify)
 
+(autoload 'sgml-lexical-context "sgml-mode")
+
 (defun sgml-mode-flyspell-verify ()
-  "Function used for `flyspell-generic-check-word-predicate' in SGML mode."
-  (not (save-excursion
-	 (let ((this (point-marker))
-	       (s (progn (beginning-of-line) (point-marker)))
-	       (e (progn (end-of-line) (point-marker))))
-	   (or (progn
-		 (goto-char this)
-		 (and (re-search-forward  "[^<]*>" e t)
-		      (= (match-beginning 0) this)))
-	       (progn
-		 (goto-char this)
-		 (and (re-search-backward "<[^>]*" s t)
-		      (= (match-end 0) this)))
-	       (and (progn
-		      (goto-char this)
-		      (and (re-search-forward  "[^&]*;" e t)
-			   (= (match-beginning 0) this)))
-		    (progn
-		      (goto-char this)
-		      (and (re-search-backward "&[^;]*" s t)
-			   (= (match-end 0) this)))))))))
+  "Function used for `flyspell-generic-check-word-predicate' in SGML mode.
+Tag and attribute names are not spell checked, everything else is.
+
+String values of attributes are checked because they can be text
+like <img alt=\"Some thing.\">."
+
+  (not (memq (car (sgml-lexical-context))
+	     '(tag pi))))
 
 ;;*---------------------------------------------------------------------*/
 ;;*    Programming mode                                                 */
@@ -384,7 +389,8 @@ property of the major mode name.")
 
 (defun flyspell-generic-progmode-verify ()
   "Used for `flyspell-generic-check-word-predicate' in programming modes."
-  (let ((f (get-text-property (point) 'face)))
+  ;; (point) is next char after the word. Must check one char before.
+  (let ((f (get-text-property (- (point) 1) 'face)))
     (memq f flyspell-prog-text-faces)))
 
 ;;;###autoload
@@ -413,8 +419,10 @@ property of the major mode name.")
 ;;*---------------------------------------------------------------------*/
 (defvar flyspell-mouse-map
   (let ((map (make-sparse-keymap)))
-    (define-key map (if (featurep 'xemacs) [button2] [down-mouse-2])
-      #'flyspell-correct-word)
+    (if (featurep 'xemacs)
+	(define-key map [button2] #'flyspell-correct-word)
+      (define-key map [down-mouse-2] #'flyspell-correct-word)
+      (define-key map [mouse-2] 'undefined))
     map)
   "Keymap for Flyspell to put on erroneous words.")
 
@@ -446,8 +454,7 @@ property of the major mode name.")
     (t (:bold t)))
   "Face used for marking a misspelled word in Flyspell."
   :group 'flyspell)
-;; backward-compatibility alias
-(put 'flyspell-incorrect-face 'face-alias 'flyspell-incorrect)
+(define-obsolete-face-alias 'flyspell-incorrect-face 'flyspell-incorrect "22.1")
 
 (defface flyspell-duplicate
   '((((class color)) (:foreground "Gold3" :bold t :underline t))
@@ -455,8 +462,7 @@ property of the major mode name.")
   "Face used for marking a misspelled word that appears twice in the buffer.
 See also `flyspell-duplicate-distance'."
   :group 'flyspell)
-;; backward-compatibility alias
-(put 'flyspell-duplicate-face 'face-alias 'flyspell-duplicate)
+(define-obsolete-face-alias 'flyspell-duplicate-face 'flyspell-duplicate "22.1")
 
 (defvar flyspell-overlay nil)
 
@@ -484,7 +490,7 @@ This runs `flyspell-mode-hook' after flyspell mode is entered or exit.
 
 Remark:
 `flyspell-mode' uses `ispell-mode'.  Thus all Ispell options are
-valid.  For instance, a personal dictionary can be used by
+valid.  For instance, a different dictionary can be used by
 invoking `ispell-change-dictionary'.
 
 Consider using the `ispell-parser' to check your text.  For instance
@@ -498,9 +504,9 @@ in your .emacs file.
   :keymap flyspell-mode-map
   :group 'flyspell
   (if flyspell-mode
-      (condition-case ()
+      (condition-case err
 	  (flyspell-mode-on)
-	(error (message "Enabling Flyspell mode gave an error")
+	(error (message "Error enabling Flyspell mode:\n%s" (cdr err))
 	       (flyspell-mode -1)))
     (flyspell-mode-off)))
 
@@ -611,7 +617,7 @@ in your .emacs file.
   ;; the welcome message
   (if (and flyspell-issue-message-flag
 	   flyspell-issue-welcome-flag
-	   (interactive-p))
+	   (called-interactively-p 'interactive))
       (let ((binding (where-is-internal 'flyspell-auto-correct-word
 					nil 'non-ascii)))
 	(message "%s"
@@ -680,6 +686,7 @@ not the very same deplacement command."
 (defvar flyspell-pre-column     nil)
 (defvar flyspell-pre-pre-buffer nil)
 (defvar flyspell-pre-pre-point  nil)
+(make-variable-buffer-local 'flyspell-pre-point)
 
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-previous-command ...                                    */
@@ -744,6 +751,9 @@ before the current command."
 		  (eq flyspell-previous-command this-command)))
 	 (or (= (current-column) 0)
 	     (= (current-column) flyspell-pre-column)
+	     ;; If other post-command-hooks change the buffer,
+	     ;; flyspell-pre-point can lie past eob (bug#468).
+	     (null (char-after flyspell-pre-point))
 	     (eq (char-syntax (char-after flyspell-pre-point)) ?w)))
     nil)
    ((not (eq (current-buffer) flyspell-pre-buffer))
@@ -926,46 +936,47 @@ Mostly we check word delimiters."
 (defun flyspell-post-command-hook ()
   "The `post-command-hook' used by flyspell to check a word in-the-fly."
   (interactive)
-  (let ((command this-command)
-	;; Prevent anything we do from affecting the mark.
-	deactivate-mark)
-    (if (flyspell-check-pre-word-p)
-	(with-current-buffer flyspell-pre-buffer
-	  '(flyspell-debug-signal-pre-word-checked)
-	  (save-excursion
-	    (goto-char flyspell-pre-point)
-	    (flyspell-word))))
-    (if (flyspell-check-word-p)
-	(progn
-	  '(flyspell-debug-signal-word-checked)
-	  (flyspell-word)
-	  ;; we remember which word we have just checked.
-	  ;; this will be used next time we will check a word
-	  ;; to compare the next current word with the word
-	  ;; that as been registered in the pre-command-hook
-	  ;; that is these variables are used within the predicate
-	  ;; FLYSPELL-CHECK-PRE-WORD-P
-	  (setq flyspell-pre-pre-buffer (current-buffer))
-	  (setq flyspell-pre-pre-point  (point)))
-      (progn
-	(setq flyspell-pre-pre-buffer nil)
-	(setq flyspell-pre-pre-point  nil)
-	;; when a word is not checked because of a delayed command
-	;; we do not disable the ispell cache.
-	(if (and (symbolp this-command) (get this-command 'flyspell-delayed))
-	    (progn
-	      (setq flyspell-word-cache-end -1)
-	      (setq flyspell-word-cache-result '_)))))
-    (while (and (not (input-pending-p)) (consp flyspell-changes))
-      (let ((start (car (car flyspell-changes)))
-	    (stop  (cdr (car flyspell-changes))))
-	(if (flyspell-check-changed-word-p start stop)
+  (when flyspell-mode
+    (let ((command this-command)
+	  ;; Prevent anything we do from affecting the mark.
+	  deactivate-mark)
+      (if (flyspell-check-pre-word-p)
+	  (with-current-buffer flyspell-pre-buffer
+	    '(flyspell-debug-signal-pre-word-checked)
 	    (save-excursion
-	      '(flyspell-debug-signal-changed-checked)
-	      (goto-char start)
-	      (flyspell-word)))
-	(setq flyspell-changes (cdr flyspell-changes))))
-    (setq flyspell-previous-command command)))
+	      (goto-char flyspell-pre-point)
+	      (flyspell-word))))
+      (if (flyspell-check-word-p)
+	  (progn
+	    '(flyspell-debug-signal-word-checked)
+	    (flyspell-word)
+	    ;; we remember which word we have just checked.
+	    ;; this will be used next time we will check a word
+	    ;; to compare the next current word with the word
+	    ;; that as been registered in the pre-command-hook
+	    ;; that is these variables are used within the predicate
+	    ;; FLYSPELL-CHECK-PRE-WORD-P
+	    (setq flyspell-pre-pre-buffer (current-buffer))
+	    (setq flyspell-pre-pre-point  (point)))
+	(progn
+	  (setq flyspell-pre-pre-buffer nil)
+	  (setq flyspell-pre-pre-point  nil)
+	  ;; when a word is not checked because of a delayed command
+	  ;; we do not disable the ispell cache.
+	  (if (and (symbolp this-command) (get this-command 'flyspell-delayed))
+	      (progn
+		(setq flyspell-word-cache-end -1)
+		(setq flyspell-word-cache-result '_)))))
+      (while (and (not (input-pending-p)) (consp flyspell-changes))
+	(let ((start (car (car flyspell-changes)))
+	      (stop  (cdr (car flyspell-changes))))
+	  (if (flyspell-check-changed-word-p start stop)
+	      (save-excursion
+		'(flyspell-debug-signal-changed-checked)
+		(goto-char start)
+		(flyspell-word)))
+	  (setq flyspell-changes (cdr flyspell-changes))))
+      (setq flyspell-previous-command command))))
 
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-notify-misspell ...                                     */
@@ -988,7 +999,7 @@ Mostly we check word delimiters."
 	  (inhibit-point-motion-hooks t)
 	  p)
       (while (and (not r) (setq p (search-backward word bound t)))
-	(let ((lw (flyspell-get-word '())))
+	(let ((lw (flyspell-get-word)))
 	  (if (and (consp lw) (string-equal (car lw) word))
 	      (setq r p)
 	    (goto-char p))))
@@ -1003,7 +1014,7 @@ Mostly we check word delimiters."
 	  (inhibit-point-motion-hooks t)
 	  p)
       (while (and (not r) (setq p (search-forward word bound t)))
-	(let ((lw (flyspell-get-word '())))
+	(let ((lw (flyspell-get-word)))
 	  (if (and (consp lw) (string-equal (car lw) word))
 	      (setq r p)
 	    (goto-char (1+ p)))))
@@ -1012,8 +1023,13 @@ Mostly we check word delimiters."
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-word ...                                                */
 ;;*---------------------------------------------------------------------*/
-(defun flyspell-word (&optional following)
-  "Spell check a word."
+(defun flyspell-word (&optional following known-misspelling)
+  "Spell check a word.
+If the optional argument FOLLOWING, or, when called interactively
+`ispell-following-word', is non-nil, checks the following (rather
+than preceding) word when the cursor is not over a word.  If
+optional argument KNOWN-MISSPELLING is non nil considers word a
+misspelling and skips redundant spell-checking step."
   (interactive (list ispell-following-word))
   (ispell-set-spellchecker-params)    ; Initialize variables and dicts alists
   (save-excursion
@@ -1038,12 +1054,14 @@ Mostly we check word delimiters."
 			  (not (memq (char-after (1- start)) '(?\} ?\\)))))
 		 flyspell-mark-duplications-flag
 		 (not (catch 'exception
-			(dolist (except flyspell-mark-duplications-exceptions)
-			  (and (string= (or ispell-local-dictionary
-					    ispell-dictionary)
-					(car except))
-			       (member (downcase word) (cdr except))
-			       (throw 'exception t)))))
+			(let ((dict (or ispell-local-dictionary
+					ispell-dictionary)))
+			  (dolist (except flyspell-mark-duplications-exceptions)
+			    (and (or (null (car except))
+				     (and (stringp dict)
+					  (string-match (car except) dict)))
+				 (member (downcase word) (cdr except))
+				 (throw 'exception t))))))
 		 (save-excursion
 		   (goto-char start)
 		   (let* ((bound
@@ -1074,29 +1092,35 @@ Mostly we check word delimiters."
 	    (setq flyspell-word-cache-end end)
 	    (setq flyspell-word-cache-word word)
 	    ;; now check spelling of word.
-	    (ispell-send-string "%\n")
-	    ;; put in verbose mode
-	    (ispell-send-string (concat "^" word "\n"))
-	    ;; we mark the ispell process so it can be killed
-	    ;; when emacs is exited without query
-	    (set-process-query-on-exit-flag ispell-process nil)
-	    ;; Wait until ispell has processed word.  Since this code is often
-            ;; executed from post-command-hook but the ispell process may not
-            ;; be responsive, it's important to make sure we re-enable C-g.
-	    (with-local-quit
-	      (while (progn
-		       (accept-process-output ispell-process)
-		       (not (string= "" (car ispell-filter))))))
-	    ;; (ispell-send-string "!\n")
-	    ;; back to terse mode.
-	    ;; Remove leading empty element
-	    (setq ispell-filter (cdr ispell-filter))
-	    ;; ispell process should return something after word is sent.
-	    ;; Tag word as valid (i.e., skip) otherwise
-	    (or ispell-filter
-		(setq ispell-filter '(*)))
-	    (if (consp ispell-filter)
-		(setq poss (ispell-parse-output (car ispell-filter))))
+            (if (not known-misspelling)
+                (progn
+                  (ispell-send-string "%\n")
+                  ;; put in verbose mode
+                  (ispell-send-string (concat "^" word "\n"))
+                  ;; we mark the ispell process so it can be killed
+                  ;; when emacs is exited without query
+                  (set-process-query-on-exit-flag ispell-process nil)
+                  ;; Wait until ispell has processed word.  Since this
+                  ;; code is often executed from post-command-hook but
+                  ;; the ispell process may not be responsive, it's
+                  ;; important to make sure we re-enable C-g.
+                  (with-local-quit
+                    (while (progn
+                             (accept-process-output ispell-process)
+                             (not (string= "" (car ispell-filter))))))
+                  ;; (ispell-send-string "!\n")
+                  ;; back to terse mode.
+                  ;; Remove leading empty element
+                  (setq ispell-filter (cdr ispell-filter))
+                  ;; ispell process should return something after word is sent.
+                  ;; Tag word as valid (i.e., skip) otherwise
+                  (or ispell-filter
+                      (setq ispell-filter '(*)))
+                  (if (consp ispell-filter)
+                      (setq poss (ispell-parse-output (car ispell-filter)))))
+              ;; Else, this was a known misspelling to begin with, and
+              ;; we should forge an ispell return value.
+              (setq poss (list word 0 '() '())))
 	    (let ((res (cond ((eq poss t)
 			      ;; correct
 			      (setq flyspell-word-cache-result t)
@@ -1151,10 +1175,15 @@ Mostly we check word delimiters."
 			      nil)
 			     (t
 			      (setq flyspell-word-cache-result nil)
-			      ;; incorrect highlight the location
+			      ;; Highlight the location as incorrect,
+			      ;; including offset specified in POSS.
 			      (if flyspell-highlight-flag
 				  (flyspell-highlight-incorrect-region
-				   start end poss)
+				   (if (and (consp poss)
+					    (integerp (nth 1 poss)))
+				       (+ start (nth 1 poss) -1)
+				     start)
+				   end poss)
 				(flyspell-notify-misspell word poss))
 			      nil))))
 	      ;; return to original location
@@ -1162,36 +1191,20 @@ Mostly we check word delimiters."
 	      (if ispell-quit (setq ispell-quit nil))
 	      res))))))))
 
-;;*---------------------------------------------------------------------*/
-;;*    flyspell-tex-math-initialized ...                                */
-;;*---------------------------------------------------------------------*/
-(defvar flyspell-tex-math-initialized nil)
-
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-math-tex-command-p ...                                  */
 ;;*    -------------------------------------------------------------    */
-;;*    This function uses the texmathp package to check if (point)      */
-;;*    is within a tex command. In order to avoid using                 */
-;;*    condition-case each time we use the variable                     */
-;;*    flyspell-tex-math-initialized to make a special case the first   */
-;;*    time that function is called.                                    */
+;;*    This function uses the texmathp package to check if point        */
+;;*    is within a TeX math environment. `texmathp' can yield errors    */
+;;*    if the document is currently not valid TeX syntax.               */
 ;;*---------------------------------------------------------------------*/
 (defun flyspell-math-tex-command-p ()
   (when (fboundp 'texmathp)
-    (cond
-     (flyspell-check-tex-math-command
-      nil)
-     ((eq flyspell-tex-math-initialized t)
-      (texmathp))
-     ((eq flyspell-tex-math-initialized 'error)
-      nil)
-     (t
-      (setq flyspell-tex-math-initialized t)
+    (if flyspell-check-tex-math-command
+        nil
       (condition-case nil
           (texmathp)
-        (error (progn
-                 (setq flyspell-tex-math-initialized 'error)
-                 nil)))))))
+        (error nil)))))
 
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-tex-command-p ...                                       */
@@ -1268,15 +1281,12 @@ this function changes the last char of the `ispell-casechars' string."
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-get-word ...                                            */
 ;;*---------------------------------------------------------------------*/
-(defun flyspell-get-word (following &optional extra-otherchars)
+(defun flyspell-get-word (&optional following extra-otherchars)
   "Return the word for spell-checking according to Ispell syntax.
-If optional argument FOLLOWING is non-nil or if `flyspell-following-word'
-is non-nil when called interactively, then the following word
-\(rather than preceding\) is checked when the cursor is not over a word.
-Optional second argument contains otherchars that can be included in word
-many times.
-
-Word syntax described by `flyspell-dictionary-alist' (which see)."
+Optional argument FOLLOWING non-nil means to get the following
+\(rather than preceding) word when the cursor is not over a word.
+Optional second argument EXTRA-OTHERCHARS is a regexp of characters
+that may be included as part of a word (see `ispell-dictionary-alist')."
   (let* ((flyspell-casechars (flyspell-get-casechars))
 	 (flyspell-not-casechars (flyspell-get-not-casechars))
 	 (ispell-otherchars (ispell-get-otherchars))
@@ -1412,7 +1422,7 @@ The buffer to mark them in is `flyspell-large-region-buffer'."
 			      ;; Move back into the match
 			      ;; so flyspell-get-word will find it.
 			      (forward-char -1)
-			      (flyspell-get-word nil)))
+			      (flyspell-get-word)))
 			   (found (car found-list))
 			   (found-length (length found))
 			   (misspell-length (length word)))
@@ -1443,7 +1453,7 @@ The buffer to mark them in is `flyspell-large-region-buffer'."
 					t
 				      nil))))
 			(setq keep nil)
-			(flyspell-word)
+			(flyspell-word nil t)
 			;; Search for next misspelled word will begin from
 			;; end of last validated match.
 			(setq buffer-scan-pos (point))))
@@ -1475,7 +1485,7 @@ The buffer to mark them in is `flyspell-large-region-buffer'."
       (goto-char (point-min))
       ;; Localwords parsing copied from ispell.el.
       (while (search-forward ispell-words-keyword nil t)
-	(let ((end (save-excursion (end-of-line) (point)))
+	(let ((end (point-at-eol))
 	      string)
 	  ;; buffer-local words separated by a space, and can contain
 	  ;; any character other than a space.  Not rigorous enough.
@@ -1551,14 +1561,17 @@ The buffer to mark them in is `flyspell-large-region-buffer'."
 			      (expand-file-name
 			       ispell-current-personal-dictionary)))))
       (setq args (append args ispell-extra-args))
-      (if (and ispell-really-aspell
-	       ispell-aspell-supports-utf8)
+
+      ;; If we are using recent aspell or hunspell, make sure we use the right encoding
+      ;; for communication. ispell or older aspell/hunspell does not support this
+      (if ispell-encoding8-command
 	  (setq args
 		(append args
 			(list
-			 (concat "--encoding="
+			 (concat ispell-encoding8-command
 				 (symbol-name
 				  encoding))))))
+
       (let ((process-coding-system-alist (list (cons "\\.*" encoding))))
 	(setq c (apply 'ispell-call-process-region beg
 		       end
@@ -1575,7 +1588,7 @@ The buffer to mark them in is `flyspell-large-region-buffer'."
 	      (flyspell-delete-region-overlays beg end)
 	      (flyspell-check-region-doublons beg end))
 	    (flyspell-external-point-words))
-	(error "Can't check region...")))))
+	(error "Can't check region")))))
 
 ;;*---------------------------------------------------------------------*/
 ;;*    flyspell-region ...                                              */
@@ -1824,7 +1837,9 @@ misspelled words backwards."
 			(throw 'exit t)))))))
 	(save-excursion
 	  (goto-char pos)
-	  (ispell-word))
+	  (ispell-word)
+	  (setq flyspell-word-cache-word nil) ;; Force flyspell-word re-check
+	  (flyspell-word))
       (error "No word to correct before point"))))
 
 ;;*---------------------------------------------------------------------*/
@@ -1902,7 +1917,7 @@ This command proposes various successive corrections for the current word."
 	  (flyspell-ajust-cursor-point pos (point) old-max)
 	  (setq flyspell-auto-correct-pos (point)))
       ;; fetch the word to be checked
-      (let ((word (flyspell-get-word nil)))
+      (let ((word (flyspell-get-word)))
 	(if (consp word)
 	    (let ((start (car (cdr word)))
 		  (end (car (cdr (cdr word))))
@@ -1994,7 +2009,7 @@ Sets `flyspell-auto-correct-previous-pos' to nil"
 ;;*    flyspell-auto-correct-previous-word ...                          */
 ;;*---------------------------------------------------------------------*/
 (defun flyspell-auto-correct-previous-word (position)
-  "Auto correct the first mispelled word that occurs before point.
+  "Auto correct the first misspelled word that occurs before point.
 But don't look beyond what's visible on the screen."
   (interactive "d")
 
@@ -2062,9 +2077,9 @@ If OPOINT is non-nil, restore point there after adjusting it for replacement."
     (error "Pop-up menus do not work on this terminal"))
   ;; use the correct dictionary
   (flyspell-accept-buffer-local-defs)
-  (or opoint (setq opoint (point-marker)))
+  (or opoint (setq opoint (point)))
   (let ((cursor-location (point))
-	(word (flyspell-get-word nil)))
+	(word (flyspell-get-word)))
     (if (consp word)
 	(let ((start (car (cdr word)))
 	      (end (car (cdr (cdr word))))
@@ -2359,5 +2374,4 @@ This function is meant to be added to `flyspell-incorrect-hook'."
 
 (provide 'flyspell)
 
-;; arch-tag: 05d915b9-e9cf-44fb-9137-fc28f5eaab2a
 ;;; flyspell.el ends here
